{
ZWECK       : Demo Netzwerkkommunikation ber MailSlots

DATEIEN     : ms01.dpr
              ms01u1.pas
              ms01u1.dfm

STAND       : 11.4.97

AUTOR       : Kurt Spitzley, email 0225438-0001@t-online.de

COPYRIGHT   : (c) 1997 bei Kurt Spitzley, All rights reserved.
              Diese Software darf zu nichtkommerziellen Zwecken frei genutzt
              und weitergegeben werden. Kommerzielle Nutzung nur mit ausdrck-
              licher Genehmigung des Autors.
GEWHR-
LEISTUNG    : Sie akzeptieren diese Software wie sie ist, ohne eine Garantie
              jeglicher Art, einschlielich aber nicht ausschlielich der
              Eignung fr eine beliebige Anwendung.

BESCHREIBUNG:

Sollen mehrere Programme (auch auf verschiedene Rechner verteilt) miteinander
kommunizieren, so kann dies ber sogenannte 'Mailslots' erfolgen. Dieses Verfahren
erlaubt das Senden von Nachrichten an einen oder mehrere Empfnger (broadcasting).
Ein Empfnger wird spezifiziert durch
-  Name des Rechners, der Domne oder '*' oder '.'
-  Name des Mailslots
Beispiele eines solchen Adressstrings sind:

\\MyGrp\mailslot\xyz      :  Domne 'MyGrp'
                             alle Applikationen mit Mailslot 'xyz'
\\MyPC\mailslot\ChannelA  :  Rechner 'MyPC'
                             alle Applikationen mit Mailslot 'ChannelA'
\\MyPC\mailslot\ChannelA  :  Rechner 'MyPC'
                             alle Applikationen mit Mailslot 'ChannelA'
\\.\mailslot\MyMailSl     : lokaler Mailslot
                            alle Applikationen mit Mailslot 'MyMailSl'
\\*\mailslot\ChannelA     : alle Rechner der Primrdomne
                            alle Applikationen mit Mailslot 'ChannelA'

Unbedingt zu beachten ist, da die Nachricht, die an einen bestimmten Mailslot
gesendet werden soll, ber jedes verfgbare Netzwerkprotokoll gesendet wird.
Dies bedeutet, da auf einem Rechner mit bsw. den Protokollen NET-BEUI, IPX und
TCP/IP eine gesendete Nachricht gleich dreimal beim Empfnger aufluft. Um hier
Konflikte zu vermeiden, sollte jede Nachricht mit einer ID versehen werden, so
da mehrfache Nachrichten ausgesondert werden knnen.

================================================================================

SCOPE       : Demo Network-Communication via MailSlots

FILES       : ms01.dpr
              ms01u1.pas
              ms01u1.dfm

LAST CHANGES: 11.4.97

AUTHOR      : Kurt Spitzley, email 0225438-0001@t-online.de

COPYRIGHT   : (c) 1997 by Kurt Spitzley, All rights reserved.
              This software should not be SOLD by anyone. It is distributed as
              freeware and therefore may be used free of charge.

DISCLAIMER  : You accept this software AS IS without any representation or
              warranty of any kind, including but not limited to the warranty of
              merchantability or fitness for a particular purpose.

DESCRIPTION:

A means for communication of multiple applications (on multiple machines too) are
'mailslots'. By this way messages can be sent to one or more recipients(broadcasting).
A recipient is specified through:
-  name of machine, domain or '*' oder '.'
-  name of mailslot
Examples for such address-strings are:

\\MyGrp\mailslot\xyz      :  domain 'MyGrp'
                             all applications with mailslot 'xyz'
\\MyPC\mailslot\ChannelA  :  machine 'MyPC'
                             all applications with mailslot 'ChannelA'
\\MyPC\mailslot\ChannelA  :  machine 'MyPC'
                             all applications with mailslot 'ChannelA'
\\.\mailslot\MyMailSl     :  local mailslot
                             all applications with mailslot 'MyMailSl'
\\*\mailslot\ChannelA     :  all machines of primary domain
                             all applications with mailslot 'ChannelA'

You should realize, that each message to be sent to a specific mailslot will be
transmitted via each available network-protocol. For a machine with e.g. the
implemented protocols NET-BEUI, IPX and TCP/IP each message will be transmitted
three times. To avoid conflicts here, each message should be signed with an
unique ID, so that at the recipients site multiple messages can be recognized.
}

unit ms01u1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Debwin, StdCtrls, ExtCtrls;

type
  EMailSlotError = class( Exception );

  TfrmMailslotDemo = class(TForm)
    gbPutIntoMailslot: TGroupBox;
    edPutIntoMailslot: TEdit;
    btnPutIntoMailslot: TButton;
    gbReadMailslot: TGroupBox;
    lboxReadMailslot: TListBox;
    btnReadMailslot: TButton;
    lblTxMailToSend: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnPutIntoMailslotClick(Sender: TObject);
    procedure btnReadMailslotClick(Sender: TObject);
  private
    MsgID: word;
    hMailServer: THandle;
    hMailSender: THandle;
  public
    { Public-Deklarationen }
  end;

var
  frmMailslotDemo: TfrmMailslotDemo;

implementation
       
{$R *.DFM}

procedure TfrmMailslotDemo.FormCreate(Sender: TObject);
begin
  MsgID:=0;

  // Create a mailslot for receiving messages

  hMailServer:=CreateMailslot('\\.\mailslot\MyMailSl',
                      0,
                      0,
                      PSecurityAttributes(nil));
  if hMailServer=INVALID_HANDLE_VALUE then
    raise EMailSlotError.Create('Error creating Mailslot');

  // Create the mailslot file handle for sending messages
                                                             
  hMailSender:=CreateFile('\\*\mailslot\MyMailSl',
                  GENERIC_WRITE,
                  FILE_SHARE_READ,
                  PSecurityAttributes(nil),
                  OPEN_EXISTING,
                  FILE_ATTRIBUTE_NORMAL,
                  0);
  if hMailSender=INVALID_HANDLE_VALUE then
    raise EMailSlotError.Create('Error creating Mailslot File');
end;

procedure TfrmMailslotDemo.FormDestroy(Sender: TObject);
begin
  CloseHandle(hMailServer);
  CloseHandle(hMailSender);
end;

procedure TfrmMailslotDemo.btnPutIntoMailslotClick(Sender: TObject);
var
  Status: boolean;
  TextToSend: string;
  NumBytesWritten: dword;
begin
  // Write message to mailslot
  inc(MsgID);
  TextToSend:=format('Mail-ID: %d Mail: %s',[MsgID,edPutIntoMailslot.Text]);
  Status:=WriteFile(hMailSender,
                    pchar(TextToSend)^,
                    length(TextToSend),
                    NumBytesWritten,
                    nil);
end;
         
procedure TfrmMailslotDemo.btnReadMailslotClick(Sender: TObject);
var
  Status: boolean;
  NextSize,
  MessageCount: dword;
  CharsRead:dword;
  s:string;
begin
  Status:=GetMailslotInfo(hMailServer,nil,NextSize,addr(MessageCount),nil);
  if NextSize<>-1 then
    begin
      SetLength(s,NextSize); // Allocate space
      ReadFile(hMailServer,PChar(s)^,NextSize,CharsRead,nil);
      SetLength(s,CharsRead); // Adjust length
    end
  else
    s:='nothing in mailslot';
  lboxReadMailSlot.Items.Add(s);
end;

end.

